#include "Music.h"
//#include "globals.h"
//#include "Sample.h"

#include <iostream>
#include <iomanip>
#include <cstring>
#include <cmath>
#include <sstream>
#include <algorithm>


#define skipSpaces				\
while (pos < text.size() && isspace(text[pos]))	\
{						\
	if (text[pos] == '\n')			\
		line++;				\
	pos++;					\
}				

#define append(value) data[channel].push_back(value)

#define error(str) {				\
	printError(str, false, name, line);	\
	return; }				\

static unsigned int pos = 0;
static int line = 1, channel = 0, prevChannel = 0, octave = 4, prevNoteLength = -1, defaultNoteLength = 8;
static int instrument[9];
static bool inDefineBlock;

static unsigned int prevLoop = -1;
static int i, j;
static bool noMusic[8][2];
static int q[9];
static bool updateQ[9];
static bool hasIntro = false;
static bool doesntLoop = false;
static bool triplet = false;
static bool inPitchSlide = false;
static bool passedIntro[8];

static int songTargetProgram;	// 0 = indeterminate/unknown/AMK, 1 = AM4, 2 = AMM.
static int targetAMKVersion;

static int loopLabel = 0;

static const int tmpTrans[19] = {0, 0, 5, 0, 0, 0, 0, 0, 0, -5, 6, 0, -5, 0, 0, 8, 0, 0, 0};
static const int instrToSample[30] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x07, 0x08, 0x09, 0x05, 0x0A,	// \ Instruments
				      0x0B, 0x01, 0x10, 0x0C, 0x0D, 0x12, 0x0C, 0x11, 0x01,		// /
				      0x00, 0x00,							// Nothing
				      0x0F, 0x06, 0x06, 0x0E, 0x0E, 0x0B, 0x0B, 0x0B, 0x0E};		// Percussion

static const int hexLengths[] = {2,2,3,4,4,1,
				 2,3,2,3,2,3,2,2,3,4,2,4,4,3,2,4,
				 1,4,4,3,2,9,3,4,2,2,3,2};
static int transposeMap[256];
static bool htranspose[256];
static int hexLeft = 0;
static int currentHex = 0;

static int tempLoopLength;		// How long the current [ ] loop is.
static int e6LoopLength;		// How long the current $E6 loop is.
static int previousLoopLength;		// How long the last encountered loop was.
static bool inE6Loop;			// Whether or not we're in an $E6 loop.
static int loopNestLevel;		// How deep we're "loop nested".
					// If this is 0, then any type of loop is allowed.
					// If this is 1 and we're in a normal loop, then normal loops are disallowed and $E6 loops are allowed.
					// If this is 1 and we're in an $E6 loop, then $E6 loops are disallowed and normal loops are allowed.
					// If this is 2, then no new loops are allowed.

static unsigned int lengths[8];		// How long each channel is.

static unsigned int tempo;		
static bool onlyHadOneTempo;

static bool sortReplacements;
//static int hTranspose = 0;00


bool sortFunction(const std::pair<const std::string, std::string> *s1, const std::pair<const std::string, std::string> *s2)
{
	return (s1->first.length() > s2->first.length());
}

bool Music::doReplacement()
{
	if (sortReplacements)
	{
		sortedReplacements.clear();
		for (auto it = replacements.begin(); it != replacements.end(); it++)
		{
			sortedReplacements.push_back(&*it);
		}

		std::sort(sortedReplacements.begin(), sortedReplacements.end(), sortFunction);
		sortReplacements = false;
	}

	for (unsigned int z = 0; z < sortedReplacements.size(); z++)
	{
		if (strncmp(text.c_str() + pos, sortedReplacements[z]->first.c_str(), sortedReplacements[z]->first.length()) == 0)
		{
			text.replace(text.begin() + pos, text.begin() + pos + sortedReplacements[z]->first.length(), sortedReplacements[z]->second.begin(), sortedReplacements[z]->second.end());
		}
	}
	return true;
}

Music::Music()
{
	playOnce = false;
	hasIntro = false;
	totalSize = 0;
	spaceForPointersAndInstrs = 0;
	exists = false;
	echoBufferSize = 0;

	//if (validateHex)		// Allow space for the buffer reservation header.
	//	data[0].resize(3);
}

void Music::init()
{
	openTextFile((std::string)"music/" + name, text);
	sortReplacements = true;
	hasYoshiDrums = false;
	onlyHadOneTempo = true;
	tempo = -1;
	guessLength = true;
	tempLoopLength = 0;
	e6LoopLength = 0;
	previousLoopLength = 0;
	inE6Loop = false;
	seconds = 0;

	songTargetProgram = 0;
	//hTranspose = 0;
	pos = 0;
	line = 1;
	channel = 0;
	octave = 4;
	prevNoteLength = -1;
	defaultNoteLength = 8;

	prevLoop = -1;
	i = 0;
	j = 0;

	inDefineBlock = false;

	for (int z = 0; z < 8; z++)
	{
		noMusic[z][0] = false;
		noMusic[z][1] = false;
		passedIntro[z] = false;
		phrasePointers[z][0] = 0;
		phrasePointers[z][1] = 0;
		lengths[z] = 0;
	}

	for (int z = 0; z < 9; z++)
	{
		q[z] = 0x7F;
		instrument[z] = 0;
		updateQ[z] = true;
	}

	hasIntro = false;
	doesntLoop = false;
	triplet = false;
	inPitchSlide = false;

	loopLabel = 0;
	currentHex = 0;
	hexLeft = 0;

	for (int z = 0; z < 256; z++)
	{
		transposeMap[z] = htranspose[z] = 0;
		usedSamples[z] = false;
	}

	for (int z = 0; z < 19; z++)
		transposeMap[z] = tmpTrans[z];

	for (int z = 0; z < 16; z++)	// Add some spaces to the end.
		text += ' ';

	for (int z = 0; z < 0x10000; z++)
		loopPointers[z] = ~0;

	if (text[0] == 0xEF && text[1] == 0xBB && text[2] == 0xBF)
	{
		text.erase(text.begin(), text.begin() + 3);
	}

	unsigned int p = text.find(";title=");
	if (p != -1)
	{
		//name.clear();
		p+=7;
		while (text[p] != '\r' && text[p] != '\n' && p < text.length())
		{
			title += text[p++];
		}
	}
	else
	{
		p = name.find_last_of('.');
		if (p != -1)
			title = name.substr(0, p);
	}


	p = text.find(";am4=");
	if (p != -1)
		songTargetProgram = 1;

	p = text.find(";amm=");
	if (p != -1)
		songTargetProgram = 2;

	if (songTargetProgram == 0)
	{
		p = text.find(";amk=");
		if (p != -1)
		{
			// Get the AMK version this song was made for.  Currently unused, but may be in the future if any syntax changes.
			p+=5;
			std::string tmp;
			while (text[p] != '\r' && text[p] != '\n' && p < text.length())
			{
				tmp += text[p++];
			}
			if (sscanf(tmp.c_str(), "%d", &targetAMKVersion) == 0) targetAMKVersion = -1;
			writeTextFile((std::string)"music/" + name, text);
		}
		else
		{
			targetAMKVersion = 0;
			if (text.find('\r') != -1)
				text = text.insert(text.length(), "\r\n\r\n;amk=1\r\n");		// Automatically assume that this is a song written for AMK.
			else
				text = text.insert(text.length(), "\n\n;amk=1\n");
			writeTextFile((std::string)"music/" + name, text);
		}
	}

	pos = 0;

	preprocess(text, name);

	text += "                       ";

	if (validateHex)			// We can't just insert this at the end due to looping complications and such.
	{
		if (text.find("#0") != -1)
			data[0].resize(3), resizedChannel = 0;
		else if (text.find("#1") != -1)
			data[1].resize(3), resizedChannel = 1;
		else if (text.find("#2") != -1)
			data[2].resize(3), resizedChannel = 2;
		else if (text.find("#3") != -1)
			data[3].resize(3), resizedChannel = 3;
		else if (text.find("#4") != -1)
			data[4].resize(3), resizedChannel = 4;
		else if (text.find("#5") != -1)
			data[5].resize(3), resizedChannel = 5;
		else if (text.find("#6") != -1)
			data[6].resize(3), resizedChannel = 6;
		else if (text.find("#7") != -1)
			data[7].resize(3), resizedChannel = 7;
	}
	else
		resizedChannel = -1;
	pos = 0;

}

void Music::compile()
{
	init();
	while (pos < text.length())
	{
		if (hexLeft != 0 && !isspace(tolower(text[pos])) && tolower(text[pos]) != '$' && text[pos] != '\n')
		{
			if (currentHex == 0xE6)
			{
				data[channel][data[channel].size() - 1] = 0xE5;
				append(0);
				append(0);
				append(0);
				hexLeft = 0;
			}
			else
			{
				error("Unknown hex command.");
			}
		}

		doReplacement();	

		switch (tolower(text[pos]))
		{
		case '?': parseQMarkDirective();	break;
		case '!': parseExMarkDirective();	break;
		case '#': parseChannelDirective();	break;
		case 'l': parseLDirective();		break;
		case 'w': parseGlobalVolumeCommand();	break;
		case 'v': parseVolumeCommand();		break;
		case 'q': parseQuantizationCommand();	break;
		case 'y': parsePanCommand();		break;
		case '/': parseIntroDirective();	break;
		case 't': parseT();			break;
		case 'o': parseOctaveDirective();	break;
		case '@': parseInstrumentCommand();	break;
		case '(': parseOpenParenCommand();	break;
		case '[': parseLoopCommand();		break;
		case ']': parseLoopEndCommand();	break;
		case '*': parseStarLoopCommand();	break;
		case 'p': parseVibratoCommand();	break;
		case '{': parseTripletOpenDirective();	break;
		case '}': parseTripletCloseDirective();	break;
		case '>': parseRaiseOctaveDirective();	break;
		case '<': parseLowerOctaveDirective();	break;
		case '&': parsePitchSlideCommand();	break;
		case '$': parseHexCommand();		break;
		case 'h': parseHDirective();		break;
		case '\"':parseReplacementDirective();	break;
		case '\n': pos++; line++;		break;
		case '|': pos++; hexLeft = 0;		break;
		case 'c': case 'd': case 'e': case 'f': case 'g': case 'a': case 'b': case 'r': case '^': 
			parseNote();			break;
		default:
			if (isspace(text[pos])) 
				{ pos++; break; }
			else 
				{ std::cout << "File " << name << ", line " << line << ": Unexpected character \"" << text[pos] << "\" found." << std::endl; pos++; break;}
		}
	}

	pointersFirstPass();
}


void Music::printChannelDataNonVerbose(int totalSize)
{
	int n = 64 - printf("%s total size", name.c_str());
	putchar(' ');
	for (int i = 0; i < n; i++)
		putchar('.');
	putchar(' ');

	printf("0x%04X bytes\n", totalSize);

}

void Music::parseQMarkDirective()
{
	pos++;
	i = getInt();
	if (i == -1) i = 0;
	switch (i)
	{
	case 0: doesntLoop = true; break;
	case 1: noMusic[channel][0] = true; break;
	case 2: noMusic[channel][1] = true; break;
	}
}
void Music::parseExMarkDirective()
{
	pos = -1;
}
void Music::parseChannelDirective()
{
	pos++;
	if (isalpha(text[pos])) 
	{ 
		parseSpecialDirective(); 
		return; 
	}

	i = getInt();
	if (i == -1) error("Error parsing octave directive.")
	if (i < 0 || i > 7) error("Illegal value for octave directive.")

	channel = i;
	q[8] = q[channel];
	updateQ[8] = updateQ[channel];
	prevNoteLength = -1;


	for (int u = 0; u < 18; u++)
	{
		if (htranspose[u])			// Undo what the h directive did.
			transposeMap[u] = tmpTrans[u];
	}
	//hTranspose = 0;
}
void Music::parseLDirective()
{
	pos++;
	i = getInt();
	
	if (i == -1) error("Error parsing \"l\" directive.")
	if (i < 1 || i > 192) error("Illegal value for \"l\" directive.")

	defaultNoteLength = i;
}
void Music::parseGlobalVolumeCommand()
{
	pos++;
	i = getInt();
	if (i == -1) error("Error parsing global volume (\"w\") command.")
	if (i < 0 || i > 255) error("Illegal value for global volume (\"w\") command.")

	append(0xE0);
	append(i);
}
void Music::parseVolumeCommand()
{
	pos++;
	i = getInt();

	if (i == -1) error("Error parsing volume (\"v\") command.")
	if (i < 0 || i > 255) error("Illegal value for global volume (\"v\") command.")

	append(0xE7);
	append(i);
}
void Music::parseQuantizationCommand()
{
	pos++;
	i = getHex();
	if (i == -1) error("Error parsing quantization (\"q\") command.")
	if (i < 1 || i > 0x7F) error("Error parsing quantization (\"q\") command.")

	q[channel] = i;
	q[8] = i;
	updateQ[channel] = true;
	updateQ[8] = true;
}
void Music::parsePanCommand()
{
	pos++;
	i = getInt();
	int pan = i;

	if (i == -1) error("Error parsing pan (\"y\") command.")
	if (i < 0 || i > 20) error("Illegal value for pan (\"y\") command.")

	skipSpaces;

	if (text[pos] == ',')
	{
		pos++;
		i = getInt();
		if (i == -1) error("Error parsing pan (\"y\") command.")
		if (i > 2)  error("Illegal value for pan (\"y\") command.")
		pan |= (i << 7);
		skipSpaces;
		if (text[pos] != ',') error("Error parsing pan (\"y\") command.")

		pos++;
		i = getInt();
		if (i == -1) error("Error parsing pan (\"y\") command.")
		if (i > 2)  error("Illegal value for pan (\"y\") command.")
		pan |= (i << 6);
	}

	append(0xDB);
	append(pan);
}
void Music::parseIntroDirective()
{
	if (channel == 8) error("Intro directive found within a loop.")
	hasIntro = true;
	pos++;
	phrasePointers[channel][1] = data[channel].size();
	prevNoteLength = -1;
	hasIntro = true;
	introLength = lengths[channel];
}
void Music::parseT()
{
	pos++;
	if (strncmp(&text[pos], "uning[", 6) == 0)
		parseTransposeDirective();
	else
		parseTempoCommand();
}
void Music::parseTempoCommand()
{
	i = getInt();
	if (i == -1) error("Error parsing tempo (\"t\") command.")
	if (i < 0 || i > 255) error("Illegal value for tempo (\"t\") command.")
	if (tempo != -1)
		onlyHadOneTempo = false;

	tempo = i;

	append(0xE2);
	append(i);

}
void Music::parseTransposeDirective()
{
	pos+=6;

	i = getInt();

	if (i == -1) error("Error parsing tuning directive.")
	if (i < 0 || i > 255)  error("Illegal instrument value for tuning directive.")

	if (text[pos] != ']') error("Error parsing tuning directive.")
	pos++;
	skipSpaces;

	if (text[pos] != '=') error("Error parsing tuning directive.")
	pos++;

	while (true)
	{
		skipSpaces;

		bool plus = true;
		if (text[pos] == '+') pos++;
		else if (text[pos] == '-') { pos++; plus = false; }

		j = getInt();

		if (j == -1) error("Error parsing tuning directive.")

		if (plus == false) j = -j;
		transposeMap[i] = j;

		skipSpaces;

		if (text[pos] != ',') break;
		pos++;
		i++;
		if (i >= 256) error("Illegal value for tuning directive.")
	}
	

}
void Music::parseOctaveDirective()
{
	pos++;
	i = getInt();
	if (i == -1) error("Error parsing octave (\"o\") directive.")
	if (i < 0 || i > 6) error("Error parsing octave (\"o\") directive.")
	octave = i;
}
void Music::parseInstrumentCommand()
{
	pos++;
	bool direct = false;
	if (text[pos] == '@') {pos++; direct = true;}

	i = getInt();
	if (i == -1) error("Error parsing instrument (\"@\") command.")
	if (i < 0 || i > 255) error("Illegal value for instrument (\"@\") command.")

	if ((i <= 18 || direct) || i >= 30)
	{
		if (convert)			
		{					
			if (i >= 0x13 && i < 30)	// Change it to an HFD custom instrument.
				i = i - 0x13 + 30;
		}
		if (optimizeSampleUsage)
			usedSamples[instrToSample[i]] = true;
		append(0xDA);
		append(i);
	}

	if (i < 30)
		if (optimizeSampleUsage)
			usedSamples[instrToSample[i]] = true;

	//hTranspose = 0;
	instrument[channel] = i;
	if (htranspose[i] == true)
		transposeMap[instrument[channel]] = ::tmpTrans[instrument[channel]];
}

void Music::parseOpenParenCommand()
{
	if (text[pos+1] == '"' || text[pos+1] == '@')
		parseSampleLoadCommand();
	else
		parseLabelLoopCommand();
}

void Music::parseSampleLoadCommand()
{
	pos+=1;
	if (text[pos] == '@')
	{
		pos++;
		i = getInt();
		i = instrToSample[i];
		if (text[pos] != ',')
		{
			error("Error parsing sample load command.")
			return;
		}
		pos++;
	}
	else
	{
		pos++;
		std::string s = "";
		while (text[pos] != '"')
		{
			if (pos >= text.length()) error("Error parsing sample load command.")
			s += text[pos];
			pos++;
		}
		pos++;
		if (text[pos] != ',')
		{
			error("Error parsing sample load command.")
			return;
		}
		i = -1;
		int gs = getSample(s);
		for (j = 0; j < mySamples.size(); j++)
		{
			if (mySamples[j] == gs)
			{
				i = j;
				break;
			}
		}

		if (i == -1) 
			error("The specified sample was not included in this song.");

		pos++;
	}
	skipSpaces;
	if (text[pos] != '$')
	{
		error("Error parsing sample load command.")
		return;
	}
	pos++;
	j = getHex();
	if (text[pos] != ')')
	{
		error("Error parsing sample load command.")
		return;
	}
	pos++;

	if (optimizeSampleUsage)
		usedSamples[i] = true;

	append(0xF3);
	append(i);
	append(j);

	
}

void Music::parseLabelLoopCommand()
{

	pos++;
	if (channel == 8) error("Nested loops are not allowed.")
	i = getInt();
	if (i == -1) error("Error parsing label loop.")
	i++;						// Needed to allow for songs that used label 0.
	if (i == 0) error("Illegal value for loop label.")
	if (i >= 0x10000)  error("Illegal value for loop label.")

	if (text[pos] != ')')  error("Error parsing label loop.")

	pos++;

	if (text[pos] == '[')				// If this is a loop definition...
		loopLabel = i, tempLoopLength = 0;	// Just set the loop label to this
	else						// Otherwise, if this is a loop call...
	{


		prevNoteLength = -1;

		if ((unsigned short)loopPointers[loopLabel] == -1) error("Label not yet defined.");
		j = getInt();
		if (j == -1) j = 1;
		if (j < 1 || j > 255)
		{
			error("Invalid loop count.")
			j = 1;
		}

		lengths[channel] += loopLengths[i] * j;
		previousLoopLength = loopLengths[i];

		append(0xE9);
		loopLocations[channel].push_back(data[channel].size());
	
		append(loopPointers[i] & 0xFF);
		append(loopPointers[i] >> 8);
		append(j);

	}

}
void Music::parseLoopCommand()
{
	pos++;

	if (text[pos] == '[')			// This is an $E6 loop.
	{
		pos++;
		if (inE6Loop == true)
			error("You cannot nest a subloop within another subloop.")
		if (loopLabel > 0 && text[pos-2] == ')')
			error("A label loop cannot define a subloop.  Use a standard loop instead.")

		inE6Loop = true;
		e6LoopLength = 0;

		append(0xE6);
		append(0x00);
		return;
	}
	else if (channel == 8)
	{				
		error("You cannot nest standard [ ] loops.");
	}


	prevLoop = data[8].size();


	prevChannel = channel;				// We're in a loop now, which is represented as channel 8.
	channel = 8;					// So we have to back up the current channel.
	prevNoteLength = -1;
	instrument[8] = instrument[prevChannel];

	if (loopLabel > 0)
	{
		if ((signed short)loopPointers[loopLabel] != -1)
		{
			error("Label redefinition.");
		}
	}

	if (loopLabel > 0)
		loopPointers[loopLabel] = prevLoop;

	tempLoopLength = 0;

}

void Music::parseLoopEndCommand()
{
	pos++;
	if (text[pos] == ']')
	{
		pos++;
		if (inE6Loop == false)
			error("A subloop end was found outside of a subloop.")
		i = getInt();
		if (i == -1)
			error("Error parsing subloop command.")
		//if (i == 0)
		//	error("A subloop cannot loop 0 times.")

		inE6Loop = false;

		if (loopLabel)
			::tempLoopLength += e6LoopLength * i;
		else
			lengths[channel] += e6LoopLength * i;
		e6LoopLength = 0;

		append(0xE6);
		append(i-1);
		return;
	}
	else if (channel != 8) error("Loop end found outside of a loop.");
	i = getInt();
	if (i == -1) i = 1;

	if (i < 1 || i > 255) error("Invalid loop count.");

	append(0);
	channel = prevChannel;

	previousLoopLength = tempLoopLength;
	lengths[channel] += tempLoopLength * i;
	if (loopLabel > 0)
		loopLengths[loopLabel] = tempLoopLength;

	append(0xE9);
	loopLocations[channel].push_back(data[channel].size());
	append(prevLoop & 0xFF);
	append(prevLoop >> 8);
	append(i);
	loopLabel = 0;
}
void Music::parseStarLoopCommand()
{
	pos++;

	if (channel == 8)  error("Nested loops are not allowed.")


	prevNoteLength = -1;
	i = getInt();
	if (i == -1) i = 1;
	if (i < 1 || i > 255)
	{
		error("Invalid loop count.");
		i = 1;
	}

	lengths[channel] += previousLoopLength * i;

	append(0xE9);
	loopLocations[channel].push_back(data[channel].size());
	append(prevLoop & 0xFF);
	append(prevLoop >> 8);
	append(i);
}
void Music::parseVibratoCommand()
{
	pos++;
	int t1, t2, t3;
	t1 = getInt();
	if (t1 == -1) error("Error parsing vibrato command.");
	skipSpaces;
	if (text[pos] != ',') error("Error parsing vibrato command.");
	pos++;
	skipSpaces;
	t2 = getInt();
	if (t2 == -1) error("Error parsing vibrato command.");
	skipSpaces;

	if (text[pos] == ',')	// The user has specified the delay.
	{
		pos++;
		skipSpaces;
		t3 = getInt();
		if (t3 == -1) error("Error parsing vibrato command.");
		if (t1 < 0 || t1 > 255) error("Illegal value for vibrato delay.")
		if (t2 < 0 || t2 > 255) error("Illegal value for vibrato rate.");
		if (t3 < 0 || t3 > 255) error("Illegal value for vibrato extent.");

		append(0xDE);
		append(t1);
		append(t2);
		append(t3);
	}
	else			// The user has not specified the delay.
	{
		if (t1 < 0 || t1 > 255) error("Illegal value for vibrato rate.");
		if (t2 < 0 || t2 > 255) error("Illegal value for vibrato extent.");

		append(0xDE);
		append(00);
		append(t1);
		append(t2);
	}

}
void Music::parseTripletOpenDirective()
{
	pos++;
	if (triplet == false) 
		triplet = true;
	else
		error("Triplet on directive found within a triplet block.");
}
void Music::parseTripletCloseDirective()
{
	pos++;
	if (triplet == true) 
		triplet = false;
	else
		error("Triplet off directive found outside of a triplet block.");
}
void Music::parseRaiseOctaveDirective()
{
	pos++;
	octave++;
	if (octave > 7)
	{
		octave = 7;
		error("The octave has been raised too high.");
	}
}
void Music::parseLowerOctaveDirective()
{
	pos++;
	octave--;
	if (octave < -1)
	{
		octave = 0;
		error("The octave has been dropped too low.");
	}
}
void Music::parsePitchSlideCommand()
{
	pos++;
	if (inPitchSlide)
	{
		error("Pitch slide directive specified multiple times in a row.");
		return;
	}
	inPitchSlide = true;
}
void Music::parseHFDInstrumentHack(int addr, int bytes)
{
	int byteNum = 0;
	do
	{
		skipSpaces;
		if (text[pos] != '$')
		{
			error("Unknown HFD hex command.");
			return;
		}
		pos++;
		i = getHex();
		if (i == -1)
		{
			error("Unknown HFD hex command.");
			return;
		}
		instrumentData.push_back(i);
		bytes--;
		byteNum++;
		addr++;
		if (byteNum == 5) 
		{
			instrumentData.push_back(0);	// On the 5th byte, we need to add a 0 as the new sub-multiplier.
			byteNum = 0;
		}

	} while (bytes >= 0);
}

void Music::parseHFDHex()
{
	skipSpaces;
	if (text[pos] == '$')
	{
		pos++;
		i = getHex();
		if (i == -1)
		{
			error("Error parsing hex command.");
			return;
		}

		if (i == 0x80 && convert)
		{
			int reg;
			int val;
			skipSpaces;
			if (text[pos] != '$')
			{
				error("Unknown HFD hex command.");
				return;
			}
			pos++;
			reg = getHex();
			if (reg == -1)
			{
				error("Unknown HFD hex command.");
				return;
			}
			skipSpaces;
			if (text[pos] != '$')
			{
				error("Unknown HFD hex command.");
				return;
			}
			pos++;
			val = getHex();
			if (val == -1)
			{
				error("Unknown HFD hex command.");
				return;
			}

			if (!(reg == 0x6D || reg == 0x7D))	// Do not write the HFD header hex bytes.
			{
				if (reg == 0x6C)			// Noise command gets special treatment.
				{
					append(0xF8);
					append(val);
				}
				else
				{
					append(0xF6);
					append(reg);
					append(val);
				}
			}
			else
			{
				songTargetProgram = 1;		// The HFD header bytes indicate this as being an AM4 song, so it gets AM4 treatment.
			}
			hexLeft = 0;
		}
		else if (i == 0x81 && convert)
		{
			skipSpaces;
			if (text[pos] != '$')
			{
				error("Unknown HFD hex command.");
				return;
			}
			pos++;
			i = getHex();
			if (i == -1)
			{
				error("Unknown HFD hex command.");
				return;
			}

			append(0xFA);
			append(0x02);
			append(i);
			hexLeft = 0;
		}
		else if ((i == 0x83) && convert)
		{
			error("Unknown HFD hex command.");
			return;
		}
		else if (i == 0x82 && convert)
		{
			int addr;
			int bytes;
			skipSpaces;
			if (text[pos] != '$')
			{
				error("Unknown HFD hex command.");
				return;
			}
			pos++;
			i = getHex();
			if (i == -1)
			{
				error("Unknown HFD hex command.");
				return;
			}
			addr = i << 8;

			skipSpaces;
			if (text[pos] != '$')
			{
				error("Unknown HFD hex command.");
				return;
			}
			pos++;
			i = getHex();
			if (i == -1)
			{
				error("Unknown HFD hex command.");
				return;
			}
			addr |= i;

			skipSpaces;
			if (text[pos] != '$')
			{
				error("Unknown HFD hex command.");
				return;
			}
			pos++;
			i = getHex();
			if (i == -1)
			{
				error("Unknown HFD hex command.");
				return;
			}
			bytes = i << 8;

			skipSpaces;
			if (text[pos] != '$')
			{
				error("Unknown HFD hex command.");
				return;
			}
			pos++;
			i = getHex();
			if (i == -1)
			{
				error("Unknown HFD hex command.");
				return;
			}
			bytes |= i;

			if (addr == 0x6136)
			{
				parseHFDInstrumentHack(addr, bytes);
				return;
			}
			do
			{
				skipSpaces;
				if (text[pos] != '$')
				{
					error("Unknown HFD hex command.");
					return;
				}
				pos++;
				i = getHex();
				if (i == -1)
				{
					error("Unknown HFD hex command.");
					return;
				}
				//append(0xF7);
				//append(i);
				//append(addr >> 16);
				//append(addr & 0xFF);	//Don't do this stuff; we won't know what we're overwriting.
				bytes--;
				addr++;

			} while (bytes >= 0);
			hexLeft = 0;

		}
		else
		{
			currentHex = 0xED;
			hexLeft = hexLengths[currentHex - 0xDA] - 1 - 1;
			append(0xED);
			if (convert)
				append(i);
			else
				append(i);
		}
	}
	else
	{
		error("Unknown HFD hex command.");
	}
}
static bool validateTremolo;
void Music::parseHexCommand()
{
	pos++;
	i = getHex();
	if (i == -1)
	{
		error("Error parsing hex command.");
		return;
	}

	if (validateHex)
	{
		if (hexLeft == 0) 
		{
			validateTremolo = true;
			currentHex = i;
			if (i > 0xFB || i < 0xDA) 
				error("Unknown hex command.");
			if (i == 0xED) 
			{
				parseHFDHex();
				return;
			}

			if (i == 0xFB)
			{
				skipSpaces;
				if (text[pos] != '$')
					error("Unknown hex command.")
				pos++;
				j = getHex();
				if (j == -1)
					error("Error parsing hex command.")
				if (j >= 0x80)
					hexLeft = 2;
				else
					hexLeft = j + 1;
				append(i);
				append(j);
				return;
			}
			else
				hexLeft = hexLengths[currentHex - 0xDA] - 1;
		}
		else
		{
			hexLeft -= 1;
			if (hexLeft == 1 && currentHex == 0xE5 && validateTremolo)
			{
				validateTremolo = false;
				if (i < 0x80)
					hexLeft++;
				else if (optimizeSampleUsage)
					usedSamples[i - 0x80] = true;
			}

			if (hexLeft == 1 && currentHex == 0xF3)
			{
				if (optimizeSampleUsage)
					usedSamples[i] = true;
			}

			if (hexLeft == 2 && currentHex == 0xF1)
			{
				echoBufferSize = max(echoBufferSize, i);
			}

			if (currentHex == 0xDA)			// If this was the instrument command
			{					// And it was >= 0x13
				if (i >= 0x13)			// Then change it to a custom instrument.
					i = i - 0x13 + 30;

				if (optimizeSampleUsage)
					usedSamples[instrumentData[(i-30)*5]] = true;
			}

			if (hexLeft == 0 && currentHex == 0xE6)
			{
				if (i == 0)
				{
					if (inE6Loop)	
						error("Cannot nest $E6 loops within other $E6 loops.");
					inE6Loop = true;
					e6LoopLength = 0;
				}
				else
				{
					if (!inE6Loop)
						error("An E6 loop starting point has not yet been declared.");
					inE6Loop = false;

					if (loopLabel)
						::tempLoopLength += e6LoopLength * i;
					else
						lengths[channel] += e6LoopLength * i;
				}
			}

			if (hexLeft == 0 && currentHex == 0xF4)
				if (i == 0x00 || i == 0x06)
					hasYoshiDrums = true;
		}



		if (i == -1)
		{
			error("Error parsing hex command.");
			return;
		}

		if (i < 0 || i > 255)
		{
			error("Illegal value for hex command.");
			return;
		}
	}
	append(i);
}
void Music::parseNote()
{
	j = text[pos];
	pos++;

	if (j == 'r') 
		i = 0xC7;
	else if (j == '^') 
		i = 0xC6;
	else 
		i = getPitch(j) - transposeMap[instrument[channel]];

	

	if (i < 0) 
		i = 0xC7;
	else if ( instrument[channel] >= 21 && instrument[channel] < 30 && i < 0xC6)
	{
		i = 0xD0 + (instrument[channel] - 21);

		if ((channel == 6 || channel == 7 || (channel == 8 && (prevChannel == 6 || prevChannel == 7)) ) == false)	// If this is not a SFX channel,
			instrument[channel] = 0xFF;										// Then don't force the drum pitch on every note.
	}

	if (inPitchSlide)
	{
		inPitchSlide = false;
		append(0xDD);
		append(0x00);
		append(prevNoteLength);
		append(i);
	}
	j = 0;

	bool okayToRewind = false;

	do
	{	
		int tempsize = j;	// If there's a pitch bend up ahead, we need to not optimize the last tie.
		int temppos = pos;	//

		if (j != 0 && (text[pos] == '^' || (i == 0xC7 && text[pos] == 'r')))
			pos++;

		j += getNoteLength(getInt());
		skipSpaces;

		if ((strncmp(text.c_str() + pos, "$DD", 3) == 0 || strncmp(text.c_str() + pos, "$dd", 3) == 0) && okayToRewind)
		{
			j = tempsize;		//
			pos = temppos;		// "Rewind" so we forcibly place a tie before the bend.
			break;			//
		}
		okayToRewind = true;

		if (pos >= text.length()) 
			break;

	} while (text[pos] == '^' || (i == 0xC7 && text[pos] == 'r'));

	if (::inE6Loop)
		e6LoopLength += j;
	else if (loopLabel != 0)		// else if because the E6 loop will add the lengths itself.
		tempLoopLength += j;
	else
		lengths[channel] += j;

	if (j >= 0x80)		// Note length must be less than 0x80
	{
		append(0x60);

		if (updateQ[channel])
		{
			append(q[channel]);
			updateQ[channel] = false;
			updateQ[8] = false;
		}
		append(i); j -= 0x60;

		while (j > 0x60)
		{
			append(0xC6); j -= 0x60;
		}

		if (j > 0)
		{
			if (j != 0x60) append(j);
			append(0xC6);
		}
		prevNoteLength = j;
		return;
	}
	else if (j > 0)
	{
		append(j);
		prevNoteLength = j;
		if (updateQ[channel])
		{
			append(q[channel]);
			updateQ[channel] = false;
			updateQ[8] = false;
		}
		append(i);
	}
	//append(i);
}
void Music::parseHDirective()
{
	pos++;
	bool negative = false;

	if (text[pos] == '-') 
	{
		negative = true;
		pos++;
	}
	i = getInt();

	if (i == -1) error("Error parsing h transpose directive.");
	if (negative) i = -i;
	transposeMap[instrument[channel]] = -i;
	htranspose[instrument[channel]] = true;
	//hTranspose = i;
}

void Music::parseSpecialDirective()
{
	if (strncmp(text.c_str() + pos, "instruments", 11) == 0 && isspace(text[pos+11]))
	{
		pos += 11;
		parseInstrumentDefinitions();

	}
	else if (strncmp(text.c_str() + pos, "samples", 7) == 0 && isspace(text[pos+7]))
	{
		pos += 7;
		parseSampleDefinitions();
		pos++;
	}
	else if (strncmp(text.c_str() + pos, "pad", 3) == 0 && isspace(text[pos+3]))
	{
		pos += 3;
		parsePadDefinition();
	}
	else if (strncmp(text.c_str() + pos, "define", 6) == 0 && isspace(text[pos+6]))
	{
		pos += 6;
		parseDefine();
	}
	else if (strncmp(text.c_str() + pos, "undef", 5) == 0 && isspace(text[pos+5]))
	{
		pos += 5;
		parseUndef();
	}
	else if (strncmp(text.c_str() + pos, "ifdef", 5) == 0 && isspace(text[pos+5]))
	{
		pos += 5;
		parseIfdef();
	}
	else if (strncmp(text.c_str() + pos, "ifndef", 6) == 0 && isspace(text[pos+6]))
	{
		pos += 6;
		parseIfndef();
	}
	else if (strncmp(text.c_str() + pos, "endif", 5) == 0 && isspace(text[pos+5]))
	{
		pos += 5;
		parseEndif();
	}
	else if (strncmp(text.c_str() + pos, "SPC", 3) == 0 && isspace(text[pos + 3]))
	{
		pos += 3;
		parseSPCInfo();
	}
	else if (strncmp(text.c_str() + pos, "louder", 6) == 0 && isspace(text[pos+6]))
	{
		pos += 6;
		parseLouderCommand();
	}
}

void Music::parseReplacementDirective()
{
	sortReplacements = true;
	pos++;

	int quotedStringLength = 0;

	std::string s = getQuotedString(text, pos, quotedStringLength);
	std::string find, replacement;

	i = s.find('=');

	if (i == -1)
		printError("Error parsing replacement directive; could not find '='", true, name, line);

	find = s.substr(0, i);
	replacement = s.substr(i+1);

	pos += quotedStringLength + 1;

	while (isspace(find[find.length() - 1]))
	{
		find.erase(find.end() - 1, find.end());
		if (find.length() == 0)
			printError("Error parsing replacement directive; string to find was of zero length.", true, name, line);
	}

	while (isspace(replacement[0]))
	{
		replacement.erase(replacement.begin(), replacement.begin()+1);
		if (replacement.length() == 0)
			printError("Error parsing replacement directive; string to replace was of zero length.", true, name, line);
	}

	replacements[find] = replacement;

	//std::sort(replacements.begin(), replacements.end(), sortFunction);
}

void Music::parseInstrumentDefinitions()
{
	enum parseState {
		lookingForOpenBrace,
		lookingForAnything,
		lookingForDollarSign,
		lookingForOpenQuote,
		gettingName,
		gettingValue,
	};

	parseState state = lookingForOpenBrace;
	
	skipSpaces;

	if (text[pos++] != '{')
		error("Could not find opening curly brace in instrument definition.");

	skipSpaces;
	while (text[pos] != '}')
	{
		if (text[pos] != '"' && text[pos] != '@'&& text[pos] != 'n')
			error("Error parsing instrument definition.");
		bool isName, isNoise;
		if (text[pos] == '"') isName = true, isNoise = false;
		else if (text[pos] == '@') isName = false, isNoise = false;
		else if (text[pos] == 'n') isName = false, isNoise = true;
		else
		{
			pos++;
			error("Error parsing instrument definition.");
		}
		pos++;
		if (isName)
		{
			std::string brrName;
			while (text[pos] != '"')
			{
				if (pos >= text.size()) error("Error parsing sample portion of the instrument definition.");
				brrName += text[pos++];
			}
			pos++;
			i = -1;
			int gs = getSample(brrName);
			for (j = 0; j < mySamples.size(); j++)
			{
				if (mySamples[j] == gs)
				{
					i = j;
					break;
				}
			}


			if (i == -1) error("The specified sample was not included in this song.")
		}
		else if (isNoise)
		{
			i = getHex();
			if (i == -1)
				error("Error parsing the noise portion of the instrument command.");

			if (i >= 0x20)
				error("Invalid noise pitch.  The value must be a hex value from 0 - 1F.")

			i |= 0x80;
		}
		else
		{
			i = getInt();
			if (i == -1)
				error("Error parsing the instrument copy portion of the instrument command.");

			if (i >= 30)
				error("Cannot use a custom instrument's sample as a base for another custom instrument.")

			i = instrToSample[i];
		}
		
		instrumentData.push_back(i);				
		if (optimizeSampleUsage)
			usedSamples[i] = true;

		skipSpaces;

		for (j = 0; j < 5; j++)
		{
			skipSpaces;
			if (text[pos++] != '$') error("Error parsing instrument definition; there were too few bytes following the sample (there must be 6).");
			i = getHex();
			if (i == -1) error("Error parsing instrument definition.");
			instrumentData.push_back(i);
		}
		skipSpaces;

	}
	pos++;
	return;

/*	//unsigned char temp;
	int count = 0;

	while (pos < text.length())
	{
		switch (state)
		{
		case lookingForOpenBrace:
			if (isspace(text[pos])) break;
			if (text[pos] != '{')
				error("Could not find opening curly brace in instrument definition.");
			state = lookingForDollarSign;
			break;
		case lookingForDollarSign:
			if (text[pos] == '\n')
				count = 0;
			if (isspace(text[pos])) break;
			if (text[pos] == '$') 
			{
				if (count == 6) error("Invalid number of arguments for instrument.  Total number of bytes must be a multiple of 6.");
				state = gettingValue;
				break;
			}
			if (text[pos] == '}')
			{
				if (count != 0) 
					error("Invalid number of arguments for instrument.  Total number of bytes must be a multiple of 6.");
				pos++;
				return;
			}

			error("Error parsing instrument definition.");
			break;
		case gettingValue:
			int val = getHex();
			if (val == -1 || val > 255) error("Error parsing instrument definition.");
			instrumentData.push_back(val);
			state = lookingForDollarSign;
			count++;
			break;
		}
		pos++;
	}*/
}

void Music::parseSampleDefinitions()
{
	skipSpaces;

	if (text[pos++] != '{')
		error("Unexpected character in sample group definition.  Expected \"{\".")


	while (text[pos] != '}')
	{
		if (pos >= text.size())
EOFTooEarly:												// Oh, laziness.
			error("Unexpected end of file found while parsing sample group definition.")

		skipSpaces;

		if (text[pos] == '\"')
		{
			pos++;
			int quotedStringLength = 0;
			std::string tempstr = getQuotedString(text, pos, quotedStringLength);
			std::string extension;
			auto tmppos = tempstr.find_last_of(".");
			if (tmppos == -1)
				error("The filename for the sample was missing its extension; is it a .brr or .bnk?")
			extension = tempstr.substr(tmppos);
			if (extension == ".bnk")
				addSampleBank((std::string)"samples/" + tempstr, this);
			else if (extension == ".brr")
				addSample((std::string)"samples/" + tempstr, this, true);
			else
				error("The filename for the sample was invalid.  Only \".brr\" and \".bnk\" are allowed.")

			pos += quotedStringLength + 1;
		}
		else if (text[pos] == '#')
		{
			pos++;
			std::string tempstr;
			while (isspace(text[pos]) == false)
			{
				if (pos >= text.size()) goto EOFTooEarly;
				tempstr += text[pos];
				pos++;
			}

			addSampleGroup(tempstr, this);
		}
		else if (text[pos] == '}')
			break;
		else if (isspace(text[pos]) == false)
		{
			error("Unexpected character found in sample group definition.");
		}

	}

}

void Music::parsePadDefinition()
{
	skipSpaces;
	if (text[pos] != '$')
		error("Error parsing padding directive.")
	pos++;
	i = getHex();
	if (i == -1)
		error("Error parsing padding directive.")

	minSize = i;

}

void Music::parseLouderCommand()
{
	append(0xF4);
	append(0x08);
}

int Music::getInt()
{
	//if (text[pos] == '$') { pos++; return getHex(); }	// Allow for things such as t$20 instead of t32.
								// Actually, can't do it.
								// Consider this:
								// l8r$ED$00$00
								// Yeah. Oh well.
	// Attempt to do a replacement.  (So things like "ab = 8"    [c1]ab    are valid).
	doReplacement();	

	int i = 0;
	int d = 0;

	while (pos < text.size() && text[pos] >= '0' && text[pos] <= '9')
	{
		d++;
		i = (i * 10) + text[pos] - '0';
		pos++;
	}

	if (d == 0) return -1; else return i;
}

int Music::getInt(const std::string &str, int &p)
{
	
	doReplacement();	

	int i = 0;
	int d = 0;

	while (p < str.size() && str[p] >= '0' && str[p] <= '9')
	{
		d++;
		i = (i * 10) + str[p] - '0';
		p++;
	}

	if (d == 0) return -1; else return i;
}

int Music::getHex()
{
	doReplacement();	


	int i = 0;
	int d = 0;
	int j;

	while (pos < text.size() && d < 2)	// AM4.05 checks for hex length, AMM doesn't.  So guess what the "d < 2" is for.
	{
		if      ('0' <= text[pos] && text[pos] <= '9') j = text[pos] - 0x30;
		else if ('A' <= text[pos] && text[pos] <= 'F') j = text[pos] - 0x37;
		else if ('a' <= text[pos] && text[pos] <= 'f') j = text[pos] - 0x57;
		else break;
		pos++;
		d++;
		i = (i*16) + j;
	}
	if (d == 0) return -1;

	return i;
}

int Music::getPitch(int i)
{
	static const int pitches[] = {9, 11, 0, 2, 4, 5, 7};

	i = pitches[i - 0x61] + (octave - 1) * 12 + 0x80;

	if (text[pos] == '+') {i++; pos++;}
	else if (text[pos] == '-') {i--; pos++;}

	if (i < 0x80)
		return -1;
	if (i >= 0xC6)
		return -2;

	return i;
}

int Music::getNoteLength(int i)
{
	//bool still = true;

	if (i == -1 && text[pos] == '=')
	{
		pos++;
		i = getInt();
		if (i == -1) //error("Error parsing note.");
		{
			printError("Error parsing note", false, name, line);
		}
		return i;
		//if (i < 1) still = false; else return i;
	}

	//if (still)
	//{
	if (i < 1 || i > 192) i = defaultNoteLength;
	i = 192 / i;

	int frac = i;

	int times = 0;
	while (pos < text.size() && text[pos] == '.')
	{
		frac = frac / 2;
		i += frac;
		pos++;
		times++;
		if (times == 2 && songTargetProgram == 1) break;	// AM4 only allows two dots for whatever reason.
	}
	//}
	if (triplet) 
		i = (int)floor(((double)i * 2.0 / 3.0) + 0.5);
	return i;
}

void Music::pointersFirstPass()
{
	if (errorCount) printError("There were errors when compiling the music file.  Compilation aborted.  Your ROM has not been modified.", true);


	if (resizedChannel != -1)
	{
		data[resizedChannel][0] = 0xFA;
		data[resizedChannel][1] = 0x04;
		data[resizedChannel][2] = echoBufferSize;
	}

	for (int z = 0; z < 8; z++)
	{
		if (data[z].size() != 0)
		{
			channel = z;
			append(0);
		}
	}



	if (mySamples.size() == 0)	// If no sample groups were provided...
	{
		text = "{#default }";		// This is a dumb, cheap trick, but...eh.
		pos = 0;
		parseSampleDefinitions();
	}

	if (game.empty())
		game = "Super Mario World (custom)";

	if (optimizeSampleUsage)
	{
		int emptySampleIndex = ::getSample("EMPTY.brr");
		if (emptySampleIndex == -1)
		{
			addSample("samples/EMPTY.brr", this, true);
			emptySampleIndex = getSample("EMPTY.brr");
		}


		for (i = 0; i < mySamples.size(); i++)
			if (usedSamples[i] == false && samples[mySamples[i]].important == false) 
				mySamples[i] = emptySampleIndex;
	}

	pos = 0;	// Pos no longer means text file position.

	if (data[0].size()) phrasePointers[0][0] = 0;
	pos = data[0].size();

	if (data[1].size()) phrasePointers[1][0] = pos;
	pos += data[1].size();

	if (data[2].size()) phrasePointers[2][0] = pos;
	pos += data[2].size();

	if (data[3].size()) phrasePointers[3][0] = pos; 
	pos += data[3].size();

	if (data[4].size()) phrasePointers[4][0] = pos;
	pos += data[4].size();

	if (data[5].size()) phrasePointers[5][0] = pos; 
	pos += data[5].size();

	if (data[6].size()) phrasePointers[6][0] = pos;
	pos += data[6].size();

	if (data[7].size()) phrasePointers[7][0] = pos;

	for (i = 0; i < 8; i++)
		 phrasePointers[i][1] += phrasePointers[i][0];

	playOnce = doesntLoop;

	spaceForPointersAndInstrs = 20;

	if (hasIntro)
		spaceForPointersAndInstrs += 18;
	if (!doesntLoop)
		spaceForPointersAndInstrs += 2;

	spaceForPointersAndInstrs += instrumentData.size();

	allPointersAndInstrs.resize(spaceForPointersAndInstrs);// = alloc(spaceForPointersAndInstrs);
	//for (i = 0; i < spaceForPointers; i++) allPointers[i] = 0x55;
	
	int add = (hasIntro ? 2 : 0) + (doesntLoop ? 0 : 2) + 4;

	//memcpy(allPointersAndInstrs.data() + add, instrumentData.base, instrumentData.size());
	for (i = 0; i < instrumentData.size(); i++)
		allPointersAndInstrs[i + add] = instrumentData[i];

	allPointersAndInstrs[0] = (add + instrumentData.size()) & 0xFF;
	allPointersAndInstrs[1] = ((add + instrumentData.size()) >> 8) & 0xFF;

	if (doesntLoop) 
	{
		allPointersAndInstrs[add - 2] = 0xFF;	// Will be re-evaluated to 0000 when the pointers are adjusted later.
		allPointersAndInstrs[add - 1] = 0xFF;
	}
	else
	{
		allPointersAndInstrs[add - 4] = 0xFE;	// Will be re-evaluated to FF00 when the pointers are adjusted later.
		allPointersAndInstrs[add - 3] = 0xFF;
		if (hasIntro)
		{
			allPointersAndInstrs[add - 2] = 0xFD;	// Will be re-evaluated to 0002 + ARAMPos when the pointers are adjusted later.
			allPointersAndInstrs[add - 1] = 0xFF;
		}
		else
		{
			allPointersAndInstrs[add - 2] = 0xFC;	// Will be re-evaluated to ARAMPos when the pointers are adjusted later.
			allPointersAndInstrs[add - 1] = 0xFF;
		}
	}
	if (hasIntro)
	{
		allPointersAndInstrs[2] = (add + instrumentData.size() + 16) & 0xFF;
		allPointersAndInstrs[3] = (add + instrumentData.size() + 16) >> 8;
	}

	add += instrumentData.size();
	allPointersAndInstrs[0+add] = data[0].size() != 0 ? (phrasePointers[0][0] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
	allPointersAndInstrs[1+add] = data[0].size() != 0 ? (phrasePointers[0][0] + spaceForPointersAndInstrs) >> 8 : 0xFF;
	allPointersAndInstrs[2+add] = data[1].size() != 0 ? (phrasePointers[1][0] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
	allPointersAndInstrs[3+add] = data[1].size() != 0 ? (phrasePointers[1][0] + spaceForPointersAndInstrs) >> 8 : 0xFF;
	allPointersAndInstrs[4+add] = data[2].size() != 0 ? (phrasePointers[2][0] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
	allPointersAndInstrs[5+add] = data[2].size() != 0 ? (phrasePointers[2][0] + spaceForPointersAndInstrs) >> 8 : 0xFF;
	allPointersAndInstrs[6+add] = data[3].size() != 0 ? (phrasePointers[3][0] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
	allPointersAndInstrs[7+add] = data[3].size() != 0 ? (phrasePointers[3][0] + spaceForPointersAndInstrs) >> 8 : 0xFF;
	allPointersAndInstrs[8+add] = data[4].size() != 0 ? (phrasePointers[4][0] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
	allPointersAndInstrs[9+add] = data[4].size() != 0 ? (phrasePointers[4][0] + spaceForPointersAndInstrs) >> 8 : 0xFF;
	allPointersAndInstrs[10+add] = data[5].size() != 0 ? (phrasePointers[5][0] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
	allPointersAndInstrs[11+add] = data[5].size() != 0 ? (phrasePointers[5][0] + spaceForPointersAndInstrs) >> 8 : 0xFF;
	allPointersAndInstrs[12+add] = data[6].size() != 0 ? (phrasePointers[6][0] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
	allPointersAndInstrs[13+add] = data[6].size() != 0 ? (phrasePointers[6][0] + spaceForPointersAndInstrs) >> 8 : 0xFF;
	allPointersAndInstrs[14+add] = data[7].size() != 0 ? (phrasePointers[7][0] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
	allPointersAndInstrs[15+add] = data[7].size() != 0 ? (phrasePointers[7][0] + spaceForPointersAndInstrs) >> 8 : 0xFF;

	if (hasIntro)
	{
		allPointersAndInstrs[16+add] = data[0].size() != 0 ? (phrasePointers[0][1] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
		allPointersAndInstrs[17+add] = data[0].size() != 0 ? (phrasePointers[0][1] + spaceForPointersAndInstrs) >> 8 : 0xFF;
		allPointersAndInstrs[18+add] = data[1].size() != 0 ? (phrasePointers[1][1] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
		allPointersAndInstrs[19+add] = data[1].size() != 0 ? (phrasePointers[1][1] + spaceForPointersAndInstrs) >> 8 : 0xFF;
		allPointersAndInstrs[20+add] = data[2].size() != 0 ? (phrasePointers[2][1] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
		allPointersAndInstrs[21+add] = data[2].size() != 0 ? (phrasePointers[2][1] + spaceForPointersAndInstrs) >> 8 : 0xFF;
		allPointersAndInstrs[22+add] = data[3].size() != 0 ? (phrasePointers[3][1] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
		allPointersAndInstrs[23+add] = data[3].size() != 0 ? (phrasePointers[3][1] + spaceForPointersAndInstrs) >> 8 : 0xFF;
		allPointersAndInstrs[24+add] = data[4].size() != 0 ? (phrasePointers[4][1] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
		allPointersAndInstrs[25+add] = data[4].size() != 0 ? (phrasePointers[4][1] + spaceForPointersAndInstrs) >> 8 : 0xFF;
		allPointersAndInstrs[26+add] = data[5].size() != 0 ? (phrasePointers[5][1] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
		allPointersAndInstrs[27+add] = data[5].size() != 0 ? (phrasePointers[5][1] + spaceForPointersAndInstrs) >> 8 : 0xFF;
		allPointersAndInstrs[28+add] = data[6].size() != 0 ? (phrasePointers[6][1] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
		allPointersAndInstrs[29+add] = data[6].size() != 0 ? (phrasePointers[6][1] + spaceForPointersAndInstrs) >> 8 : 0xFF;
		allPointersAndInstrs[30+add] = data[7].size() != 0 ? (phrasePointers[7][1] + spaceForPointersAndInstrs) & 0xFF : 0xFB;
		allPointersAndInstrs[31+add] = data[7].size() != 0 ? (phrasePointers[7][1] + spaceForPointersAndInstrs) >> 8 : 0xFF;
	}

	totalSize = data[0].size() + data[1].size() + data[2].size() + data[3].size() + data[4].size() + data[5].size() + data[6].size() + data[7].size() + data[8].size() + spaceForPointersAndInstrs;
	

	if (verbose) 
		std::cout << name << " total size: 0x" << std::hex << std::uppercase << std::setw(4) << std::setfill('0') << totalSize << " bytes" << std::dec << std::endl;
	else	     
		printChannelDataNonVerbose(totalSize);
	//for (int z = 0; z <= 8; z++)
	//{
		if (verbose) 
		{
			printf("\t#0: 0x%03X #1: 0x%03X #2: 0x%03X #3: 0x%03X Ptrs+Instrs: 0x%03X\n\t#4: 0x%03X #5: 0x%03X #6: 0x%03X #7: 0x%03X Loop:        0x%03X \n", data[0].size(), data[1].size(), data[2].size(), data[3].size(), spaceForPointersAndInstrs, data[4].size(), data[5].size(), data[6].size(), data[7].size(), data[8].size());
			int spaceUsedBySamples = 0;
			for (i = 0; i < mySamples.size(); i++)
			{
				spaceUsedBySamples += 4 + samples[mySamples[i]].data.size();	// The 4 accounts for the space used by the SRCN table.
			}
			printf("Space used by echo: 0x%04X bytes.  Space used by samples: 0x%04X bytes.\n\n", echoBufferSize << 11, spaceUsedBySamples);
		}
		//}
	if (totalSize > minSize && minSize > 0)
		std::cout << "File " << name << ", line " << line << ": Warning: Song was larger than it could pad out by 0x" << std::hex << std::uppercase << std::setw(4) << std::setfill('0') << totalSize - minSize << " bytes." << std::dec << std::endl;


	mainLength = -1;
	for (i = 0; i < 7; i++)
		if (lengths[i] != 0)
			mainLength = min(mainLength, lengths[i]);
	
	if (hasIntro)
		mainLength -= introLength;

	if (tempo == -1) tempo = 0x36;

	if (guessLength)
		seconds = (unsigned int)std::floor((mainLength * 2 + introLength) / (2.0 * tempo) + 0.5);	// Just 2? Not 2.012584 or something?  Wow.
}

void Music::parseDefine()
{
	error("A #define was found after the preprocessing stage.");
	//skipSpaces;
	//std::string defineName;
	//while (!isspace(text[pos]) && pos < text.length())
	//{
	//	defineName += text[pos++];
	//}

	//for (unsigned int z = 0; z < defineStrings.size(); z++)
	//	if (defineStrings[z] == defineName)
	//		error("A string cannot be defined more than once.");

	//defineStrings.push_back(defineName);
}

void Music::parseUndef()
{
	error("An #undef was found after the preprocessing stage.");
//	skipSpaces;
//	std::string defineName;
//	while (!isspace(text[pos]) && pos < text.length())
//	{
//		defineName += text[pos++];
//	}
//	unsigned int z = -1;
//	for (z = 0; z < defineStrings.size(); z++)
//		if (defineStrings[z] == defineName)
//			goto found;
//	
//	error("The specified string was never defined.");
//
//found:
//
//	defineStrings[z].clear();
}

void Music::parseIfdef()
{	
	error("An #ifdef was found after the preprocessing stage.");
//	inDefineBlock = true;
//	skipSpaces;
//	std::string defineName;
//	while (!isspace(text[pos]) && pos < text.length())
//	{
//		defineName += text[pos++];
//	}
//
//	unsigned int z = -1;
//
//	int temp;
//
//	for (unsigned int z = 0; z < defineStrings.size(); z++)
//		if (defineStrings[z] == defineName)
//			goto found;
//
//	temp = text.find("#endif", pos);
//
//	if (temp == -1)
//		error("#ifdef was missing a matching #endif.");
//
//	pos = temp;
//found:
//	return;
}

void Music::parseIfndef()
{
	error("An #ifndef was found after the preprocessing stage.");
//	inDefineBlock = true;
//	skipSpaces;
//	std::string defineName;
//	while (!isspace(text[pos]) && pos < text.length())
//	{
//		defineName += text[pos++];
//	}
//
//	unsigned int z = -1;
//
//	for (unsigned int z = 0; z < defineStrings.size(); z++)
//		if (defineStrings[z] == defineName)
//			goto found;
//
//	return;
//
//found:
//	int temp = text.find("#endif", pos);
//
//	if (temp == -1)
//		error("#ifdef was missing a matching #endif.");
//
//	pos = temp;
	

}

void Music::parseEndif()
{
	error("An #endif was found after the preprocessing stage.");
	//if (inDefineBlock == false)
	//	error("#endif was found without a matching #ifdef or #ifndef")
	//else
	//	inDefineBlock = false;
}

void Music::parseSPCInfo()
{
	skipSpaces;
	if (text[pos] != '{')
		error("Could not find opening brace in SPC info command.");

	pos++;
	skipSpaces;

	while (text[pos] != '}')
	{
		if (text[pos] != '#')
			error("Unexpected symbol found in SPC info command.  Expected \'#\'.");
		pos++;
		std::string typeName;

		while (!isspace(text[pos]))
			typeName += text[pos++];

		if (typeName != "author" && typeName != "comment" && typeName != "title" && typeName != "game" && typeName != "length")
		{
			error("Unexpected type name found in SPC info command.  Only \"author\", \"comment\", \"title\", \"game\", and \"length\" are allowed.");
		}

		skipSpaces;

		if (text[pos] != '\"')
			error("Unexpected symbol found in SPC info command.  Expected a quoted string.");

		pos++;
		int quotedStringLength = 0;
		std::string parameter = getQuotedString(text, pos, quotedStringLength);

		if (typeName == "author")
			author = parameter;
		else if (typeName == "comment")
			comment = parameter;
		else if (typeName == "title")
			title = parameter;
		else if (typeName == "game")
			game = parameter;
		else if (typeName == "length")
		{
			if (parameter == "auto")
			{
				guessLength = true;
			}
			else
			{
				guessLength = false;
				int p = 0;
				i = getInt(parameter, p);
				if (parameter[p] != ':' || i == -1)
					error("Error parsing SPC length field.  Format must be m:ss or \"auto\".")

				p++;
				j = getInt(parameter, p);
				if (j == -1 || p != parameter.length())
					error("Error parsing SPC length field.  Format must be m:ss or \"auto\".")

				seconds = i * 60 + j;

				if (seconds > 0xFFFFFF)
					error("Congratulations on creating an absurdly long piece of music, but songs longer than 11 days are not allowed by the SPC format.")

				seconds = seconds & 0xFFFFFF;


			}
		}

		pos += quotedStringLength + 1;

		skipSpaces;
	}

	if (author.length() > 32)
	{
		author = author.substr(0, 32);
		printWarning((std::string)("\"Author\" field was too long.  Truncating to \"") + author + "\".");
	}
	if (game.length() > 32)
	{
		game = game.substr(0, 32);
		printWarning((std::string)("\"Game\" field was too long.  Truncating to \"") + game + "\".");
	}
	if (comment.length() > 32)
	{
		comment = comment.substr(0, 32);
		printWarning((std::string)("\"Comment\" field was too long.  Truncating to \"") + comment + "\".");
	}
	if (title.length() > 32)
	{
		title = title.substr(0, 32);
		printWarning((std::string)("\"Title\" field was too long.  Truncating to \"") + title + "\".");
	}

	pos++;
}
