#ifndef _MUSIC_H
#define _MUSIC_H
#include <vector>
#include <string>
#include <map>
#include "globals.h"

class Music
{


public:
	std::string name;
	std::vector<byte> data[9];
	std::vector<unsigned short> loopLocations[8];
	bool playOnce;
	bool hasIntro;
	unsigned short phrasePointers[8][2];
	unsigned short loopPointers[0x10000];
	unsigned int loopLengths[0x10000];		// How long, in ticks, each loop is.
	std::string text;
	int totalSize;
	int spaceForPointersAndInstrs;
	std::vector<byte> allPointersAndInstrs;
	std::vector<byte> instrumentData;
	std::vector<byte> finalData;
	
	unsigned int introLength;
	unsigned int mainLength;

	unsigned int seconds;

	bool hasYoshiDrums;


	int index;

	//byte mySamples[255];
	std::vector<unsigned short> mySamples;
	//int mySampleCount;
	int echoBufferSize;

	std::string title;
	std::string author;
	std::string game;
	std::string comment;

	bool usedSamples[255];		// Holds a record of which samples have been used for this song.


	int minSize;

	bool exists;

	int posInARAM;

	void compile();

	std::map<std::string, std::string> replacements;
	std::vector<const std::pair<const std::string, std::string> *> sortedReplacements;
	Music();

	void init();
	bool doReplacement();
private:
	void pointersFirstPass();
	void parseQMarkDirective();
	void parseExMarkDirective();
	void parseChannelDirective();
	void parseLDirective();
	void parseGlobalVolumeCommand();
	void parseVolumeCommand();
	void parseQuantizationCommand();
	void parsePanCommand();
	void parseIntroDirective();
	void parseT();
	void parseTempoCommand();
	void parseTransposeDirective();
	void parseOctaveDirective();
	void parseInstrumentCommand();
	void parseOpenParenCommand();
	void parseLabelLoopCommand();
	void parseSampleLoadCommand();
	void parseLoopCommand();
	void parseLoopEndCommand();
	void parseStarLoopCommand();
	void parseVibratoCommand();
	void parseTripletOpenDirective();
	void parseTripletCloseDirective();
	void parseRaiseOctaveDirective();
	void parseLowerOctaveDirective();
	void parsePitchSlideCommand();
	void parseHexCommand();
	void parseNote();
	void parseHDirective();
	void parseReplacementDirective();
	
	void parseSpecialDirective();
	void parseInstrumentDefinitions();
	void parseSampleDefinitions();
	void parsePadDefinition();
	void parseASMCommand();
	void parseJSRCommand();
	void parseLouderCommand();
	void compileASM();

	void parseDefine();
	void parseIfdef();
	void parseIfndef();
	void parseEndif();
	void parseUndef();

	void parseSPCInfo();

	
	//std::vector<std::string> defineStrings;

	void printChannelDataNonVerbose(int);
	void parseHFDHex();
	void parseHFDInstrumentHack(int addr, int bytes);
	void insertedZippedSamples(const std::string &path);

	int getInt();
	int getInt(const std::string &str, int &p);
	int getHex();
	int getPitch(int j);
	int getNoteLength(int);

	bool guessLength;
	int resizedChannel;
};

#endif
